/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <sched.h>
#include <libspe2.h>
#include <mars/mars.h>

#define NUM_TASKS	6
#define NUM_THREADS	6
#define NUM_ENTRIES	100
#define QUEUE_DEPTH	50

struct queue_entry {
	char text[64];
};

extern struct spe_program_handle task_spe_prog;

static struct mars_context mars_ctx;
static struct mars_task_id task_id[NUM_TASKS];
static struct mars_task_params task_params;
static struct mars_task_args task_args;
static struct mars_task_queue task_queue;

static pthread_t thread_id[NUM_THREADS];
static int thread_arg[NUM_THREADS];

static void *thread(void *arg)
{
	int ret, i;
	struct queue_entry data __attribute__((aligned(MARS_TASK_QUEUE_ENTRY_ALIGN)));

	printf("HOST  : Thread %d - Started\n", *((int *)arg) + 1);

	for (i = 0; i < NUM_ENTRIES; i++) {
		sprintf(data.text, "Host Thread %d - Data %d", *((int *)arg) + 1, i + 1);

		ret = mars_task_queue_push(&task_queue, &data);
		if (ret) {
			printf("Host thread push failed! (%d)\n", ret);
			pthread_exit(&ret);
		}

		printf("HOST  : Thread %d - push: '%s'\n",
			*((int *)arg) + 1, data.text);

		sched_yield();
	}

	printf("HOST  : Thread %d - Finished\n", *((int *)arg) + 1);

	pthread_exit(&ret);

	return NULL;
}

int main(void)
{
	int ret, i;
	struct queue_entry buffer_task_queue[QUEUE_DEPTH] __attribute__((aligned(MARS_TASK_QUEUE_BUFFER_ALIGN)));
	
	ret = mars_initialize(&mars_ctx, NULL);
	if (ret) {
		printf("MARS context initialize failed! (%d)\n", ret);
		return 1;
	}

	ret = mars_task_queue_initialize(&mars_ctx, &task_queue, &buffer_task_queue,
				sizeof(struct queue_entry), QUEUE_DEPTH,
				MARS_TASK_QUEUE_HOST_TO_MPU);
	if (ret) {
		printf("MARS task queue initialize failed! (%d)\n", ret);
		return 1;
	}

	for (i = 0; i < NUM_TASKS; i++) {
		char name[MARS_TASK_NAME_LEN_MAX];

		snprintf(name, MARS_TASK_NAME_LEN_MAX, "Task %d", i + 1);
		task_params.name = (const char *)name;
		task_params.elf_image = task_spe_prog.elf_image;
		task_params.context_save_size = MARS_TASK_CONTEXT_SAVE_SIZE_MAX;
		ret = mars_task_initialize(&mars_ctx, &task_id[i], &task_params);
		if (ret) {
			printf("MARS task initialize failed! (%d)\n", ret);
			return 1;
		}

		task_args.type.u64[0] = (uint64_t)(uintptr_t)&task_queue;
		task_args.type.u32[2] = (uint32_t)NUM_ENTRIES;
		ret = mars_task_schedule(&task_id[i], &task_args, 0);
		if (ret) {
			printf("MARS task schedule failed! (%d)\n", ret);
			return 1;
		}
	}

	for (i = 0; i < NUM_THREADS; i++) {
		thread_arg[i] = i;
		ret = pthread_create(&thread_id[i], NULL, thread, &thread_arg[i]);
		if (ret) {
			printf("Host thread create failed! (%d)\n", ret);
			return 1;
		}
	}

	for (i = 0; i < NUM_THREADS; i++) {
		ret = pthread_join(thread_id[i], NULL);
		if (ret) {
			printf("Host thread join failed! (%d)\n", ret);
			return 1;
		}
	}

	for (i = 0; i < NUM_TASKS; i++) {
		ret = mars_task_wait(&task_id[i]);
		if (ret) {
			printf("MARS task wait failed! (%d)\n", ret);
			return 1;
		}

		ret = mars_task_finalize(&task_id[i]);
		if (ret) {
			printf("MARS task finalize failed! (%d)\n", ret);
			return 1;
		}
	}

	ret = mars_finalize(&mars_ctx);
	if (ret) {
		printf("MARS context finalize failed! (%d)\n", ret);
		return 1;
	}

	return 0;
}


